"""
Documark categories for the Series class
"""


CATEGORIES = {
    "constructor": "Constructing new time series",
    "conversion": "Converting time series frequency",
    "information": "Getting information about time series",
    "manipulation": "Manipulating time series values",
    "comparison": "Comparing time series",
    "multiple": "Combining multiple time series",
    "homogenizing": "Homogenizing and extrapolating time series",
    "filtering": "Filtering time series",
    "moving": "Applying moving window functions",
    "temporal_change": "Calculating temporal change",
    "temporal_change_conversion": "Converting measures of temporal change",
    "temporal_cumulation": "Calculating temporal cumulation",
    "indexing": "Indexing time series",
    "serialization": "Serializing and deserializing time series",
}

